package lpsolve;

import java.io.File;


import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.events.PluginLifeCycleEventListener;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.framework.util.Cast;
import org.processmining.framework.util.OsUtil;

/**
 * <p>
 * Title:
 * </p>
 * 
 * <p>
 * Description:
 * </p>
 * 
 * <p>
 * Copyright: Copyright (c) 2004
 * </p>
 * 
 * <p>
 * Company:
 * </p>
 * 
 * @author not attributable
 * @version 1.0
 */
public class ProMLpSolve {

	private static boolean loaded = false;

	public static void initialize(PluginContext context, File baseDir) {
		String path = baseDir.getAbsolutePath();
		if (OsUtil.isRunningWindows()) {
			// Run on a Windows machine
			path += System.getProperty("file.separator")
					+ "lib"//
					+ System.getProperty("file.separator") + "win"
					+ (OsUtil.is32Bit() ? "32" : "64")//
					+ System.getProperty("file.separator");

			// Load libraries required by LpSolve.
			System.load(path + "lpsolve55.dll");
			System.load(path + "lpsolve55j.dll");

		} else if (OsUtil.isRunningLinux() || OsUtil.isRunningUnix()) {
			path += System.getProperty("file.separator")
					+ "lib"//
					+ System.getProperty("file.separator")
					+ "ux"//
					+ (OsUtil.is32Bit() ? "32" : "64")
					+ System.getProperty("file.separator");

			// Load libraries required by LpSolve.
			System.load(path + "liblpsolve55.so");
			System.load(path + "liblpsolve55j.so");
		} else if (OsUtil.isRunningMacOsX()) {
			path += System.getProperty("file.separator") + "lib"//
					+ System.getProperty("file.separator") + "mac"//
					+ System.getProperty("file.separator");

			// Load libraries required by LpSolve.
			System.load(path + "liblpsolve55.dylib");
			System.load(path + "build-osx");
		}
		if (context != null) {
			context.log("Using LpSolve version "
					+ LpSolve.lpSolveVersion().getMajorversion() + "."
					+ LpSolve.lpSolveVersion().getMinorversion() + "."
					+ LpSolve.lpSolveVersion().getRelease() + "."
					+ LpSolve.lpSolveVersion().getBuild(), MessageLevel.DEBUG);
		} else {
			System.out.println("Using LpSolve version "
					+ LpSolve.lpSolveVersion().getMajorversion() + "."
					+ LpSolve.lpSolveVersion().getMinorversion() + "."
					+ LpSolve.lpSolveVersion().getRelease() + "."
					+ LpSolve.lpSolveVersion().getBuild());
		}
		loaded = true;
	}

	private ProMLpSolve() {
	}

	/**
	 * Construct and LpSolver object, and registers the context as a listener,
	 * such that it receives messages and can pass abort (cancel) events.
	 * 
	 * Note that, if the given context is not null, then the methode deleteLp()
	 * should not be called to delete the Lp from memory. Instead, use
	 * <code>ProMLpSolve.unregisterListeners(solver)</code>. This way, any
	 * listeners are nicely removed.
	 * 
	 * By default, if a context is given, the solver listens for the context to
	 * finish execution, after which the LP is automatically deleted from memory
	 * and any listeners are unregistered.
	 * 
	 * @param context
	 *            the context
	 * @param rows
	 *            the rows
	 * @param cols
	 *            the number of columns
	 * @return an LpSolver
	 * @throws LpSolveException
	 */
	public static LpSolve getSolver(PluginContext context, int rows, int cols)
			throws LpSolveException {
		if (!loaded) {
			throw new NoClassDefFoundError("LpSolver not initialized.");
		}
		final LpSolve solver = LpSolve.makeLp(rows, cols);
		if (context != null) {
			installListners(context, solver);
			context.getPluginLifeCycleEventListeners().add(
					new PluginLifeCycleEventListener() {

						public void pluginCancelled(PluginContext context) {
							try {
								unInstallListners(solver);
								context.getPluginLifeCycleEventListeners()
										.remove(this);
							} catch (LpSolveException e) {
								context.log(e);
							}
						}

						public void pluginCompleted(PluginContext context) {
							try {
								unInstallListners(solver);
								context.getPluginLifeCycleEventListeners()
										.remove(this);
							} catch (LpSolveException e) {
								context.log(e);
							}
						}

						public void pluginCreated(PluginContext context) {
							// Gracefully ignore
						}

						public void pluginDeleted(PluginContext context) {
							try {
								unInstallListners(solver);
								context.getPluginLifeCycleEventListeners()
										.remove(this);
							} catch (LpSolveException e) {
								context.log(e);
							}
						}

						public void pluginFutureCreated(PluginContext context) {
							// Gracefully ignore
						}

						public void pluginResumed(PluginContext context) {
							// Gracefully ignore
						}

						public void pluginStarted(PluginContext context) {
							// Gracefully ignore
						}

						public void pluginSuspended(PluginContext context) {
							// Gracefully ignore
						}

						public void pluginTerminatedWithError(
								PluginContext context, Throwable t) {
							try {
								unInstallListners(solver);
								context.getPluginLifeCycleEventListeners()
										.remove(this);
							} catch (LpSolveException e) {
								context.log(e);
							}
						}

					});
		}

		return solver;
	}

	/**
	 * Construct and LpSolver object.
	 * 
	 * @param rows
	 *            the rows
	 * @param cols
	 *            the number of columns
	 * @return an LpSolver
	 * @throws LpSolveException
	 */
	public static LpSolve getSolver(int rows, int cols) throws LpSolveException {
		return getSolver(null, rows, cols);
	}

	public static boolean isErrorSolveCode(int code, PluginContext context) {
		switch (code) {
		case -2: {
			context.log("Out of Memory while solving", MessageLevel.ERROR);
			break;
		}
		case 0: {
			context.log("Optimal solution found while solving",
					MessageLevel.DEBUG);
			return false;
		}
		case 1: {
			context.log("Suboptimal solution found while solving",
					MessageLevel.ERROR);
			break;
		}
		case 2: {
			context.log("The model is infeasible", MessageLevel.ERROR);
			break;
		}
		case 3: {
			context.log("The model is unbounded", MessageLevel.ERROR);
			break;
		}
		case 4: {
			context.log("The basic model is degenerative", MessageLevel.ERROR);
			break;
		}
		case 5: {
			context.log("A numerical failure occurred", MessageLevel.ERROR);
			break;
		}
		case 6: {
			context.log("Solver aborted", MessageLevel.ERROR);
			break;
		}
		case 7: {
			context.log("Solver timed-out", MessageLevel.ERROR);
			break;
		}
		case 9: {
			context.log("Solution found during preSolve", MessageLevel.DEBUG);
			return false;
		}
		case 10: {
			context.log("Branch and Bound routine failed", MessageLevel.ERROR);
			break;
		}
		case 11: {
			context.log("Branch and Bound routine stopped", MessageLevel.ERROR);
			break;
		}
		case 12: {
			context.log("Found feasible B&B solution", MessageLevel.DEBUG);
			return false;
		}
		case 13: {
			context.log("No feasible B&B solution found", MessageLevel.ERROR);
			break;
		}
		}
		// problem.printLp();
		return true;
	}

	public static void installListners(PluginContext context, LpSolve solver)
			throws LpSolveException {
		// install abort monitor
		solver.putAbortfunc(new AbortListener() {
			public boolean abortfunc(LpSolve solver, Object progress)
					throws LpSolveException {
				// the object progress can safely be cast to progress, as it is
				// defined
				// by context.getProgress();
				return Cast.<Progress> cast(progress).isCancelled();
			}
		}, context.getProgress());

		// install log monitor
		solver.putLogfunc(new LogListener() {
			public void logfunc(LpSolve s, Object context, String message)
					throws LpSolveException {
				Cast.<PluginContext> cast(context).log(message,
						MessageLevel.DEBUG);
			}
		}, context);

		// set to ignore stdOut
		solver.setOutputfile("");

		// set verbode level
		solver.setVerbose(LpSolve.SEVERE);
	}

	public static void unInstallListners(LpSolve solver)
			throws LpSolveException {
		// install abort monitor
		solver.putAbortfunc(null, null);

		// install log monitor
		solver.putLogfunc(null, null);

		// set to ignore stdOut
		solver.setOutputfile("");

		// set verbode level
		solver.setVerbose(LpSolve.SEVERE);

		solver.deleteLp();
	}

	/**
	 * isFeasibleSolution
	 * 
	 * @param solution
	 *            an array with values for all variables. index 0 is the first
	 *            variable and the size of the array equals
	 *            solver.getNcolumns();
	 * @return boolean
	 */
	public static boolean isFeasible(LpSolve solver, double[] solution)
			throws LpSolveException {
		if (solution.length != solver.getNcolumns()) {
			return false;
		}
		int[] indices = new int[solution.length];
		for (int i = 0; i < indices.length; i++) {
			indices[i] = i + 1;
		}

		boolean valid = true;
		double val = 0;
		double rh = 0;
		int type = 0;
		int row = 0;
		double[] constr = new double[solver.getNcolumns() + 1];
		for (int i = 0; valid && (i < solver.getNrows()); i++) {
			row = i + 1;
			val = solver.getConstrValue(row, indices.length, solution, indices);
			type = solver.getConstrType(row);
			rh = solver.getRh(row);
			solver.getRow(row, constr);
			if (type == LpSolve.GE) {
				// val should be >= rh
				valid = (val >= rh);
			} else if (type == LpSolve.LE) {
				// val should be <= rh
				valid = (val <= rh);
			} else { // type == problem.EQ
				valid = (val == rh);
			}
			// check for possible cancellation (return false in that case)
		}
		return valid;
	}
}
